#!/usr/bin/perl

# Helmut Sennewald 04/25/2017
# Performs A THD Analysis At A Defined Number Of Values(e.g. Frequencies)
#
#   
# Open a command window in XP. Go to the design folder.
# Run the Perl script:
# perl ltspdisto.pl input.net output.cir DEC 10 20 20000
#
#---- ADJUST THE VARIABLE $ltspice ACCORDINGLY TO THE LOCATION ON YOUR PC ----
# Examples:
# $ltspice = '"C:\programs\LTC\SwCADIII\scad3.exe" -b' ;
# $ltspice = '"C:\Programme\LTC\SwCADIII\scad3.exe" -b' ;
#
#$ltspice = '"C:\Programme\LTC\LTspiceIV\scad3.exe" -b' ;
$ltspice = '"C:\Programme\LTC\LTspiceXVII\XVIIx64.exe" -b' ;
#
# Searched string in input.net file 
# There have to be a SPICE line  .param F1=[yourNumbers]
# Exactly this line is used by this program to step the frequency parameter F1 .
$key_freq        = uc ('.param F1=') ; 
#
# Enable or disable phase of harmonic in output files
$kphaseOption    = 1 ; # 1=enabled, 0=off
#
#
#
# The number of harmonics is defined in the schematics in the .four comamnd.
#
# THD in LTspice:  voltage-harmonic / voltage-fundamental * 100%
#
# THD  : Total harmonic distortion
# AC_1 : AC amplitude at fundamental
# DC   : DC offset
# k1   : 100%
# k2..kn : distorton at n*fundamental
# p0   : 0
# p1..pn : phase of n*fundamnetal
#
# A parameter in the netlist will be modified before each simulation run.
# The results from the Error Log File are written into the file output.txt 
# and an output.raw file.
# The output.raw file can be viewed in the waveform viewer of LTspice.
#
# Output format of distortion:
# ----------------------------
#   V(THD) V(AC_1) V(DC) V(k1) V(k2) V(k3) ....... V(k100)
# Additional(optional) phase:
#   I(p0) I(p1) I(p2) .... I(p100)
#
# Usage: perl kfac2.pl input.net output.cir LIN|OCT|DEC nr start stop
#
# nr = number of steps (DEC: per decade |OCT: per octave | LIN: per range)
# start, stop = start and stop values(e.g. frequency)
# 
# Setup in the schematic or netlist:
# ----------------------------------
#
# 1 Make your design ready for the THD sweep.
#   Tools -> Control Panel -> Operation Automatically delete .net files: NO
#
# 2 Either use 'View -> Netlist' or RUN to generate an actual .net-file. 
#
# 3 Only the line ".param F1=....." will be altered from run to run.
#   For a typical THD versus frequency the value {F1} has to be used in 
#   .tran, .four and in the voltage source.
#   In principle you can step with F1 what you want in your circuit, e.g. 
#   the bias current of a transistor. This way you can plot THD versus current.
#
#  Example: typical case of THD versus ferquency
# .param F1=2k
# .param TSTOP=10/F1
# .param TSTART=5/F1
# .param TSTEP=0.001/F1
# .tran 0 {TSTOP} {TSTART} {TSTEP}
# .op
# .four {F1} V(out) 9  
# V1 IN 0 SINE(0 1 {F1})
#
#
# Output data in a text file:
# ---------------------------
#
# All output values are in %.
# Output.thd file:
#  frequency              THD            DC           k1          k2 
# 1.0000000000e+002,   16.4662320,    0.1021540,  100.0000000,   ....
# 2.1544346900e+002,   16.4661960,    0.1021540,  100.0000000,   ....
#
# If $kphaseOption=1 then there are more columns. pn = phase(kn), p0 = 0
# All phase values are in degree.
#      p0           p1            p2          p3         p4
# 0.00000000,   0.4662320,    0.1021540,  2.30000000,   ....
# 0.00000000,   1.4661960,    0.1021540,  3.30000000,   ....
#
#
# Output data in a text file:
# ---------------------------
#
# Output.raw file
# Standard  .OP parameter sweep
#
#
#
# Searched strings in .log-file 
$key_data_begin  = uc ('Number 	  [Hz]') ;
$key_data_end    = uc ('Total Harmonic') ;
$key_thd         = uc ('Total Harmonic Distortion:') ;
$key_thd_varname = uc ('Fourier components of') ;
$key_dc_value    = uc ('DC component:') ;
#
#
@kfactors = (0 .. 101) ;
@$kfactorsPhase = (0 .. 101) ;
# 
# 
#
print "THD Sweep V0.6\n";
#
#----------------------------- Command line processing  
#
# Check for source .net name
$fileSrc = $ARGV[0] ;
$p1 = index ($fileSrc, '.') ;
if ($p1 < 1) {
  print "Error: Missing '.' in file name \'$fileSrc\'\n" ;
  print "Usage: perl ltspdisto.pl source.net dest.cir LIN|OCT|DEC nr start stop\n" ;
  print "Program aborted\n" ;
  exit(1) ; }
#
# Check for .cir name
$fileIn = $ARGV[1] ;
$p1 = index ($fileIn, '.cir') ;
if ($p1 < 1) {
  print "Error: Missing '.cir' in file name \'$fileIn\'\n" ;
  print "Usage: perl ltspdisto.pl source.net dest.cir LIN|OCT|DEC nr start stop\n" ;
  print "Program aborted\n" ;
  exit(1) ; }
else {
  $file0 = substr ($fileIn, 0, $p1) ; }


# Calculate sweep steps
$sweepType  = uc ($ARGV[2]) ;
$sweepNr1   = $ARGV[3] ;
$sweepStart = $ARGV[4] ;
$sweepStop  = $ARGV[5] ;
$sweepIncr  = 0 ;

if (index($sweepType, 'DEC') >= 0) { 
  $sweepNrTotal = $sweepNr1 / log(10) * log($sweepStop/$sweepStart) ; 
  $sweepIncr = exp(log(10)/$sweepNr1) ; }
else {
  if (index($sweepType, 'OCT') >= 0) { 
    $sweepNrTotal = $sweepNr1 / log(2) * log($sweepStop/$sweepStart) ;
    $sweepIncr = exp(log(2)/$sweepNr1) ;}  
  else {
    if (index($sweepType, 'LIN') >= 0) { 
      $sweepNrTotal = $sweepNr1 ; 
      $sweepIncr = ($sweepStop - $sweepStart)/$sweepNr1 ; }  
    else {
      print 'Program terminated with error! Only  LIN|DEC|OCT  sweeps are supported.'; 
      exit(1) ; }
  }
}

$sweepNrTotal = int ($sweepNrTotal+0.5) ;
printf "Total number of simulation steps: %d \n", $sweepNrTotal+1 ;


#----------------- Main loop for simulation

# Open THD file
$fileThd = $file0 . '.thd' ;
open (FTHD, ">$fileThd") ;


# --- Main Loop ---
for ($n = 0; $n < $sweepNrTotal+1; $n++) {

# Calculate next frequency
  if ($n > 0) {
    if (index($sweepType, 'DEC') >= 0) { 
      $frequency = $frequency * $sweepIncr ; }
    else {
      if (index($sweepType, 'OCT') >= 0) { 
      $frequency = $frequency * $sweepIncr ; }
      else {
        if (index($sweepType, 'LIN') >= 0) { 
      $frequency = $frequency + $sweepIncr ; }
      }
    }
  } 
  else { $frequency = $sweepStart ; }


  # Create .cir file with new frequency
  $fileIn  = $fileSrc ;
  $fileOut = $file0 . '.cir' ;
  open (FIN, "$fileIn") ;
  open (FOUT, ">$fileOut") ;

  printf "%4d ", $n+1 ;
  print "Changing netlist file  \'$fileIn -> $fileOut\'   " ;

  while ( $line_in = <FIN> ) {
    $line_in = uc ($line_in) ;
    $p1 = index ($line_in, $key_freq) ;
    if ($p1 < 0) 
    {
      printf FOUT "$line_in" ;
    }
    else
    {
     printf      "F1=%14.10e \n", $frequency ;
     printf FOUT "$key_freq %14.10e \n", $frequency ;
    }
  }    

  close (FIN) ;
  close (FOUT) ;



#--- Call LTspice
  print "     Calling LTspice \'$fileOut\'\n" ;
  system ("$ltspice $fileOut") ;
  
  
 
#--- Read THD from Log-file 
  $fileIn  = $file0 . '.log' ;

  open (FIN, "$fileIn") ;

  print "     Read/write THD \'$fileIn -> $fileThd\'   " ;

  $dataMode = 0;
  $kimax = 0;
  while ( $line_in = <FIN> ) {
    $line_in = uc ($line_in) ;
    $p1k = index ($line_in, $key_thd) ;
    $p4k = index ($line_in, $key_thd_varname) ;
    $p8k = index ($line_in, $key_data_begin) ;
    $p9k = index ($line_in, $key_data_end) ;
    $p10 = index ($line_in, $key_dc_value) ;
  
    if ($p9k >= 0) {
      $dataMode = 0; }
    
    if ($dataMode == 1) {
      chomp($line_in) ;
#string0 string1   string2   string3     string4   string5   
#  1   	1.000e+04	8.180e-01	1.000e+00	   -3.29	    0.00
#  2   	2.000e+04	1.209e-01	1.478e-01	  -98.75	  -95.46
      string_6 ($line_in, $string0, $string1, $string2, $string3, $string4, $string5);
      $ki = 1*$string0 ;
      if ($ki == 1) { $baseAmplitude = $string2 ; }
      $kfactors[$ki] = 100*$string3 ;
      $kfactorsPhase[$ki] = $string4 ;
      if ($kimax < $ki) { $kimax = $ki ; } #highest harmonic number
    } 

    if ($p8k >= 0) { 
      $dataMode = 1; }

# Name of node, e.g. V(out)
    if ($p4k >= 0) {
      chomp($line_in) ;
      $p4  = index ($line_in, $key_thd_varname) ; 
      $p5 = $p4+length($key_thd_varname) ;
      $p6 = length($line_in) ;
      $varName = substr($line_in, $p5+1, $p6 - $p5);
    }
  
# DC component
    if ($p10 >= 0) {
      chomp($line_in) ;
      $p5 = length($key_dc_value) ;
      $p6 = length($line_in) ;
      $kfactors[0] = substr($line_in, $p5, $p6 - $p5);
    }

# Write to THD-file  
    if ($p1k >= 0) {
      $p1  = index ($line_in, $key_thd) ;
      $p2 = $p1 + length ($key_thd) + 1;
      $p3 = index ($line_in, '%') ;
      $thd = substr ($line_in, $p2, $p3-$p2) ;
      printf "   THD=%12.7f\n", $thd ;
#    
      printf FTHD "%14.10e, %12.7f, %12.7f", $frequency, $thd, $baseAmplitude ;
      for ($nn = 0; $nn <= $kimax; $nn++) {
        printf FTHD ", %12.7f", $kfactors[$nn] ; }
        $kfactorsPhase[0] = 0 ;
        if ($kphaseOption == 1) {    
          for ($nn = 0; $nn <= $kimax; $nn++) {
            printf FTHD ", %12.7f", $kfactorsPhase[$nn] ; }
        }
        printf FTHD "\n" ;
    }
  }    
  close (FIN) ;
} ; # end for loop
close (FTHD) ;



#----------------- Create a raw-file for the waveform viewer 
#
if ($kphaseOption == 1) { $kimaxph = $kimax+1 ;}
else {$kimaxph = 0 ;}                             # +1 <- kfactorsPhase[0]
#
# Read first and second line from last run
$fileIn  = $file0 . '.raw' ;

$fileInTemp  = $file0 . '.raw1' ;
system ("ltsputil17raw4.exe $fileIn $fileInTemp") ;
system ("copy /Y $fileInTemp  $fileIn") ;

open (FIN, "$fileIn") ;
$line_in = <FIN> ;
$line_in1 = <FIN> ;
close FIN ;
# Write THD .raw-file
$fileOut  = $file0 . '.raw' ;
open (FOUT, ">$fileOut") ;
print "Writing THD to .raw-file \'$fileOut\'\n" ;
printf FOUT "%s", $line_in ;
printf FOUT "%s", $line_in1 ;
#printf FOUT "%s\n", 'Title: * C:\_NEU_LTSPICE\38 perl\ac1.asc' ;
#printf FOUT "%s\n", 'Date: Sun Sep 03 00:00:00 2006' ;
printf FOUT "%s\n", 'Plotname: Operating Point' ;
printf FOUT "%s\n", 'Flags: real stepped' ;
printf FOUT "%s", 'No. Variables: ' ;
printf FOUT "%d\n", $kimax+4+$kimaxph ;
printf FOUT "%s", 'No. Points:    ' ;     
printf FOUT "  %d\n", $sweepNrTotal+1 ;
printf FOUT "%s\n", 'Offset:   0.0000000000000000e+000' ;
printf FOUT "%s\n", 'Command: Linear Technology Corporation LTspice' ;
printf FOUT "%s\n", 'Variables:' ;
printf FOUT "%s\n", '  0	 F1	 param' ;
printf FOUT "%s  %s  %s\n", '  1	', 'V(THD)', 'voltage' ;
#printf FOUT "%s  %s  %s\n", '  2	', $varName, 'voltage' ;
printf FOUT "%s  %s  %s\n", '  2	', 'V(AC_1)', 'voltage' ;
printf FOUT "%s  %s  %s\n", '  3	', 'V(DC)', 'voltage' ;
for ($n=1; $n <=$kimax; $n++) {
  printf FOUT "  %d   %s%d%s  %s\n", $n+3, 'V(k', $n, ')', 'voltage' ; }
if ($kphaseOption == 1) {  
  for ($m=0; $m <=$kimax; $m++) {
    printf FOUT "  %d   %s%d%s  %s\n", $m+3+$n, 'I(p', $m, ')', 'current' ; }
}
printf FOUT "%s\n", 'Values:' ;

# Read from THD-file
$fileIn  = $file0 . '.thd' ;
open (FIN, "$fileIn") ;
for ($n = 0; $n < $sweepNrTotal+1; $n++) {
  printf FOUT "%d", $n ;  # Step number
  $line_in = <FIN> ;
  chomp($line_in);
  $p1 = index($line_in, ',') ;
  $number1 = substr($line_in, 0, $p1) ; # THD
  printf FOUT " %14.10e\n", $number1 ;
  $line_in = substr($line_in, $p1+1, length($line_in)-$p1) ;
  
  $kimaxtot = $kimax + $kimaxph ;
  for ($m = 1; $m < $kimaxtot+4; $m++) {
    if ($m < $kimaxtot+3) {
      $p1 = index($line_in, ',') ;
      $number1 = substr($line_in, 0, $p1) ;
      $line_in = substr($line_in, $p1+1, length($line_in)-$p1) ;
    }
    else {
      $p1 = length($line_in) ;
      $number1 = substr($line_in, 0, $p1) ;
    }
    printf FOUT "  %14.10e\n", $number1 ;
  }
}
close (FOUT) ;
close (FIN) ;

# print "Conversion finished\n" ;
exit(0) ;



#----------------- Subroutines(s)
#
# return the first 6 space seperated substrings of a string
sub string_6 {
  my $line_in = $_[0] ;
  my $string0 = '' ;
  my $string1 = '' ;
  my $string2 = '' ;
  my $string3 = '' ;
  my $string4 = '' ;
  my $string5 = '' ;
  my $p1, $p2, $p3 ;
  my $strx = '' ;
  my $line_in1 = '' ;
  $p1 = 0 ;
  $p2 = 0 ;
  $p3 = 0 ;
  $p4 = length ($line_in) ;

# remove some characters
  $line_in1 = '' ;
  for ($p1=0; $p1<$p4; $p1++) {
    $strx = substr ($line_in, $p1, 1);
    if ( ($strx ge "\t") && ($strx le 'z') ) {
      $line_in1 = $line_in1 . $strx ; }
  }
# replace tabs with blanks
  $line_in = '' ;
  $p4 = length ($line_in1) ;
  for ($p1=0; $p1<$p4; $p1++) {
    $strx = substr ($line_in1, $p1, 1);
    if ($strx eq "\t") {
      $line_in = $line_in . ' ' ; } 
    else {
      $line_in = $line_in . $strx ; }
  }
 
# skip leading blancs and other characters in line
  $p4 = length ($line_in) ;
  $p1 = 0 ;
  while (substr ($line_in, $p1, 1) le ' ') {
    $p1++;
  }
  $p2 = $p1 ;

  $flag1 = 0;
  $flag2 = 0;
  $string0 = '';
  while (($p2 < $p4) && ($flag2 == 0)) {
    $char1 = substr ($line_in, $p2, 1) ;
    if ($char1 gt ' '){
      $flag1 = 1;
      $string0 = $string0 . $char1 ;
    }
    if (($char1 le ' ') && ($flag1 == 1)){
      $flag2 = 1;
    }
    $p2 = $p2 + 1 ;
  }  

  $flag1 = 0;
  $flag2 = 0;
  $string1 = '';
  while (($p2 < $p4) && ($flag2 == 0)) {
    $char1 = substr ($line_in, $p2, 1) ;
    if ($char1 gt ' '){
      $flag1 = 1;
      $string1 = $string1 . $char1 ;
    }
    if (($char1 le ' ') && ($flag1 == 1)){
      $flag2 = 1 ;
    }
    $p2 = $p2 + 1 ;
  }  

  $flag1 = 0;
  $flag2 = 0;
  $string2 = '';
  while (($p2 < $p4) && ($flag2 == 0)) {
    $char1 = substr ($line_in, $p2, 1) ;
    if ($char1 gt ' '){
      $flag1 = 1;
      $string2 = $string2 . $char1 ;
    }
    if (($char1 le ' ') && ($flag1 == 1)){
      $flag2 = 1;
    }
    $p2 = $p2 + 1 ;
  }  

  $flag1 = 0;
  $flag2 = 0;
  $string3 = '';
  while (($p2 < $p4) && ($flag2 == 0)) {
    $char1 = substr ($line_in, $p2, 1) ;
    if ($char1 gt ' '){
      $flag1 = 1;
      $string3 = $string3 . $char1 ;
    }
    if (($char1 le ' ') && ($flag1 == 1)){
      $flag2 = 1;
    }
    $p2 = $p2 + 1 ;
  }  
  
  $flag1 = 0;
  $flag2 = 0;
  $string4 = '';
  while (($p2 < $p4) && ($flag2 == 0)) {
    $char1 = substr ($line_in, $p2, 1) ;
    if ($char1 gt ' '){
      $flag1 = 1;
      $string4 = $string4 . $char1 ;
    }
    if (($char1 le ' ') && ($flag1 == 1)){
      $flag2 = 1;
    }
    $p2 = $p2 + 1 ;
  }  
  
  $flag1 = 0;
  $flag2 = 0;
  $string5 = '';
  while (($p2 < $p4) && ($flag2 == 0)) {
    $char1 = substr ($line_in, $p2, 1) ;
    if ($char1 gt ' '){
      $flag1 = 1;
      $string5 = $string5 . $char1 ;
    }
    if (($char1 le ' ') && ($flag1 == 1)){
      $flag2 = 1;
    }
    $p2 = $p2 + 1 ;
  }   
  $_[1] = $string0 ;
  $_[2] = $string1 ;
  $_[3] = $string2 ;
  $_[4] = $string3 ;
  $_[5] = $string4 ;
  $_[6] = $string5 ;
}

